################################################################################
#
#   01_stam.R
#
#   Package initialization and application interface.
#
#   written by Claudio Lottaz (COMPDIAG-MPI-MG), in August 2004
#
################################################################################
#
#   Package initialization
#

.onLoad <- function(libname, pkgname) {
  # make sure to register S4 classes
  require(methods)
  require(utils)

  # check for dot (from graphviz)
  if (.Platform$OS.type == "unix") {
    if (system("which dot > /dev/null", ignore.stderr=TRUE) != 0) {
      warning("stam did not find 'dot'. Graph layout will not work!")
    }
  }

  # initialize StAM's server feature if possible
  data(server.config, package="stam", envir=parent.env(environment()))

  # mark interactive mode
  if (is.null(getOption("interactiveStam"))) options(interactiveStam=interactive())

  packageStartupMessage("Welcome to Structured Analysis of Microarrays\n")
}

# Generic Methods
setGeneric("getGOterm", function(x,...) standardGeneric("getGOterm"))
setGeneric("getGOchildren", function(x,...) standardGeneric("getGOchildren"))
setGeneric("getGOparents", function(x,...) standardGeneric("getGOparents"))
setGeneric("writeHTML", function(x,...) standardGeneric("writeHTML"))
setGeneric("print", function(x,...) standardGeneric("print"))

setOldClass("pamrtrained")

# install an environment restricted to stam for storing intermediate results
stam.env <- new.env(parent=emptyenv())

###############################################################################
#
#   User Interface
#

stam.net <- function(chip="hgu95av2", root="GO:0008150", probes=character(0)) {
  net <- new("stamNet", chip=chip, root=root, probes=probes)
  return(net)
}

stam.cv <- function(expression.matrix,classifications,chip="hgu95av2",root="GO:0008150",
                    beta = NULL, deltas = NULL, ndeltas = 10, results.per.node=FALSE,
                    old.cv = NULL, pamimagefile = NULL, verbose = FALSE) {
  # for efficiency we wildly use side-effects on through cv in stam.env
  if (class(expression.matrix) %in% c("exprSet","ExpressionSet")) {
    if (length(classifications) == 1) 
      classifications <- pData(expression.matrix)[,classifications]
    expression.matrix <- exprs(expression.matrix)
  }
  if (is.null(colnames(expression.matrix))) 
    colnames(expression.matrix) <- 1:ncol(expression.matrix)
  if (ncol(expression.matrix) != length(classifications))
    stop("'classifications' must hold one entry per column in 'expression.matrix'")
  require(package=paste(chip, ".db", sep=""), character.only=TRUE)
  probesOnChip <- length(eval(as.symbol(paste(chip,"ACCNUM",sep=""))))
  if (nrow(expression.matrix) != probesOnChip)
    stop("Expected ", probesOnChip," probe-sets for ", chip, ", received ", nrow(expression.matrix))
  if (!is.null(old.cv)) {
    if (class(old.cv) != "stamCV") stop("'old.cv' must be of class 'stamCV'")
    stam.env[["cv"]] <- old.cv
    if (!stam.matchcv(old.cv, expression.matrix, classifications, 
                      chip, root, beta, deltas, ndeltas)) ans <- "n"
    ans <- "y"
  } else {
    ans <- "n"
    if (!is.null(pamimagefile)) {
      if (file.exists(pamimagefile)) {
        ans <- readline(paste("Use precomputed pamimage '", pamimagefile, "' [Y/n]? ", sep=""))
        if (!(ans %in% c("n", "N"))) { 
          load(pamimagefile, envir=.GlobalEnv)
          require(package=paste(chip, ".db", sep=""), character.only=TRUE)
          if (!stam.matchcv(stam.env[["cv"]], expression.matrix, classifications, 
                            chip, root, beta, deltas, ndeltas)) ans <- "n"
        }
      }
    }
  }
  if (ans %in% c("n", "N")) {
    stam.env[["cv"]] <- new("stamCV", expression.matrix, classifications, beta, chip, root)
    stam.fit.pam(expression.matrix, verbose=verbose)
    if (!is.null(pamimagefile)) save(stam.env, file=pamimagefile, compress=TRUE)
  }
  cat(" - Finding best graph shrinkage...\n")
  stam.shrink(expression.matrix, deltas, ndeltas, results.per.node) # reads and writes in stam.env[["cv"]]
  cv <- stam.env[["cv"]]
  rm("cv", envir=stam.env)
  return(cv)
}

stam.fit <- function(cv, expression.matrix, collapse.scnodes=FALSE,
                     alpha = 0.5, delta = NULL, max.nodes = 100) {
  if (class(cv) != "stamCV") stop("'cv' must be of class 'stamCV'")
  if (class(expression.matrix) %in% c("exprSet","ExpressionSet")) expression.matrix <- exprs(expression.matrix)
  if (is.null(colnames(expression.matrix))) 
    colnames(expression.matrix) <- paste("X", 1:ncol(expression.matrix), sep="")
  if (ncol(expression.matrix) != length(cv@sample.labels))
    stop("'cv' has not the same number of samples as 'expression.matrix'")
  require(package=paste(cv@chip, ".db", sep=""), character.only=TRUE)
  idx <- unlist(mget(rownames(expression.matrix), cv@probes))
  if (sum(idx != 1:length(idx)) > 0)
    stop("Probesets in 'cv' do not match rownames in 'expression.matrix'")
  cat(" - Fitting for best shrinkage...\n")
  fit <- new("stamFit", cv, expression.matrix, alpha=alpha, 
             delta=delta, max.nodes=max.nodes, collapse.scnodes=collapse.scnodes)
  return(fit)
}

stam.predict <- function(fit, expression.matrix, classifications=NULL, testset=NULL) {
  if (class(fit) != "stamFit") stop("'fit' must be of class 'stamFit'")
  if (class(expression.matrix) %in% c("exprSet","ExpressionSet")) expression.matrix <- exprs(expression.matrix)
  if (is.null(colnames(expression.matrix))) 
    colnames(expression.matrix) <- paste("X", 1:ncol(expression.matrix), sep="")
  require(package=paste(fit@chip, ".db", sep=""), character.only=TRUE)
  idx <- unlist(mget(rownames(expression.matrix), fit@probes))
  if (sum(idx != 1:length(idx)) > 0)
    stop("Probesets in 'fit' do not match rownames in 'expression.matrix'")
  if (!is.null(classifications))
    if ((ncol(expression.matrix) != length(classifications)) &&
        (ncol(expression.matrix) - length(testset) != length(classifications))) 
      stop("'classifications' must hold one entry per column in 'expression.matrix'")
  classifications <- as.character(classifications)
  if (is.null(classifications) && !is.null(testset)) 
    stop("Testset specified but no classifications given")
  if (is.null(testset)) {
    if (is.null(classifications)) testset <- numeric(0)
    else testset <- 1:ncol(expression.matrix)
  }
  if (sum(!(testset %in% 1:ncol(expression.matrix))) > 0)
    stop("Testset must contain indices in 1:ncol(expression.matrix)")
  cat(" - Predicting new samples...\n")
  prediction <- new("stamPrediction", fit, expression.matrix, classifications, testset)
  return(prediction)  
}

stam.evaluate <- function(expression.matrix, classifications, report.dir = getwd(),
                          aclass = names(table(classifications))[1], titlestem = NULL, 
                          testset = stam.balanced.folds(classifications, 3)[[1]],
                          chip = "hgu95av2", root = "GO:0008150", no.output = FALSE,
                          alpha = seq(0,1, 0.1), beta = NULL, deltas = NULL, 
                          ndeltas = 30, minspec = NULL, minsens = 0.1, maxsens = 1.0, 
                          pamimagefile=NULL) {

  # check arguments
  if (class(expression.matrix) %in% c("exprSet","ExpressionSet")) {
    if (length(classifications) == 1) 
      classifications <- pData(expression.matrix)[,classifications]
    expression.matrix <- exprs(expression.matrix)
  }
  if (is.null(colnames(expression.matrix))) 
    colnames(expression.matrix) <- paste("X", 1:ncol(expression.matrix), sep="")
  if (ncol(expression.matrix) != length(classifications) &&
      ncol(expression.matrix) - length(testset) != length(classifications))
    stop("'classifications' must hold one entry per sample or training sample")
  if (!is.null(beta)) {
    nbeta <- ifelse(is.matrix(beta), ncol(beta), length(beta))
    if (nbeta != length(table(classifications)))
      stop("'classifications' and 'beta' don't match in number of classes")
  }
  if (!(aclass %in% classifications))
    stop("'aclass' not found in 'classifications'")
  if (!is.null(testset))
    if (sum(testset %in% seq(ncol(expression.matrix))) < length(testset))
      stop("not all samples in 'testset' found in 'classifications'")
  if (length(report.dir) != 1) 
    stop("More than one report directory specified:",report.dir)
  if (file.info(report.dir)["isdir"] != TRUE) 
    stop("Report directory seems not to be a valid directory:", report.dir)

  
  # test- and trainingset, the same if testset is null
  if (is.null(testset)) {
    if (is.null(titlestem)) 
      titlestem <- paste("StAM Report: Prediction of", aclass, "(explorative)")
    trainingset <- seq(ncol(expression.matrix))
    testset <- seq(ncol(expression.matrix))
  } else {
    if (is.null(titlestem)) 
      titlestem <- paste("StAM Report: Prediction of", aclass, "(predictive)")
    trainingset <- setdiff(seq(ncol(expression.matrix)), testset)
  }

  # compute and report for first beta
  cv <- NULL
  got.testcls <- ncol(expression.matrix) == length(classifications)
  if (got.testcls) traincls <- classifications[trainingset] 
  else traincls <- classifications
  if (is.null(beta)) beta <- table(traincls)/length(traincls)
  if (!is.matrix(beta)) beta <- t(as.matrix(beta))
  else beta <- beta[order(as.character(beta[,1])),]
  pamimagefile <- pamimagefile
  if (!is.null(pamimagefile)) {  
    tryerr <- try(con <- file(pamimagefile, open="w"))
    if (class(tryerr) == "try-error") pamimagefile=NULL
    else {
      close(con)
      if (file.info(pamimagefile)$size == 0) file.remove(pamimagefile)
    }
  }
  for (i in 1:nrow(beta)) {
    curr.beta <- beta[i,]
    cv <- stam.cv(expression.matrix[,trainingset], traincls, old.cv = cv,
                  chip = chip, root = root, pamimagefile = pamimagefile,
                  beta = curr.beta, deltas = deltas, ndeltas = ndeltas)
    if (nrow(cv@results) == 0) {
      cat("No suitable shrinkage factor found\n")
      next;
    }
    fit <- stam.fit(cv, expression.matrix[,trainingset], alpha=alpha)
    pred <- stam.predict(fit, expression.matrix, classifications, testset=testset)
    eval <- new("stamEval", expression.matrix, cv, fit, pred, testset)
    if (no.output) return(eval)
    i <- 1
    repeat {
      new.dir <- paste(report.dir, "/analysis_", gsub(" ","0",formatC(i,dig=4)), sep="")
      if (!file.exists(new.dir)) break
      i <- i + 1
    }
    dir.create(new.dir)
    cat("    . writing in directory", new.dir, "...\n")
    save(eval, file=paste(new.dir, "/eval.RData", sep=""), compress=TRUE)
    curr.beta <- round(curr.beta, 3)
    stam.writeHTML(eval, outfile=new.dir, aclass = aclass,
                   minspec = minspec, minsens = minsens, maxsens = maxsens)
  }
  invisible(eval)
}  

stam.writeHTML <- function(x, title=NULL, align="left", outfile="index.html", 
                           nonodes = FALSE, ...) {
  if (missing(x)) stop("No object to report on provided")
  require(package=paste(x@chip, ".db", sep=""), character.only=TRUE)
  if (is.null(title)) {
    if (class(x) == "stamNet")        title <- "StAM-Report on Predictor Net"
    if (class(x) == "stamCV")         title <- "StAM-Report on Cross-Validation"
    if (class(x) == "stamFit")        title <- "StAM-Report on Model Fit"
    if (class(x) == "stamPrediction") title <- "StAM-Report on Prediction Results"
  }
  nocache=FALSE
  if (class(x) == "stamEval") {
    aclass <- list(...)$aclass
    if (is.null(aclass)) aclass <- x@cv@class.labels[1]
    else {
      if (!(aclass %in% x@cv@class.labels))
        stop("aclass (",aclass,") not found in stamEval object (", 
             toString(x@cv@class.labels), ")")
    }
    if (outfile == "index.html") path <- "."
    else path <- sub("/*$", "", outfile)
    curr.path <- getwd()      # expand the path
    setwd(path)
    path <- getwd()
    setwd(curr.path)
    outfile <- paste(path, "index.html", sep="/")
    nocache=TRUE
    if (is.null(title)) {
      no.testset <- identical(as.numeric(x@testset), as.numeric(1:ncol(x@exprs)))
      title <- paste("StAM Report: Prediction of", aclass,
                     ifelse(no.testset, "(explorative)", "(predictive)"),
                     "- beta", toString(round(x@cv@beta, 3)))
    }
  }
  if (is.null(title)) title <- "StAM-Report"
  beginHTML(title=title, align=align, outfile=outfile, nocache=nocache)
  writeHTML(x, nonodes=nonodes, ...)
  endHTML()

  # for evaluations write node info in separate files
  if ((class(x) == "stamEval") && !nonodes) {
    stam.env[["allnodes"]] <- x@fit@nodes
    stam.env[["GOids"]] <- new.env(hash=TRUE)
    for (i in 1:length(stam.env[["allnodes"]])) 
      stam.env[["GOids"]][[stam.env[["allnodes"]][[i]]@ID]] <- i
    if (!exists("cv", envir=stam.env)) stam.env[["cv"]] <- x@cv
    stam.env[["exprs"]] <- x@exprs
    for (i in length(x@fit@nodes):1) {
      tmp <- gsub(":", "", x@fit@nodes[[i]]@ID)
      outfile <- paste(path, "/", tmp,
                       ifelse(x@fit@nodes[[i]]@supNode,"a.html",".html"), sep="")
      beginHTML(title=x@fit@nodes[[i]]@ID, align=align, outfile=outfile, nocache=nocache)
      writeHTML(x@fit@nodes[[i]], sepfiles=TRUE)
      stam.nodePerformanceHTML(x@fit,  i, "Performance on training data")      
      stam.nodePerformanceHTML(x@pred, i, "Performance on test data")
      stam.nodePredictHTML(x@pred, i)
      endHTML()
    }
    rm("cv", "exprs", "allnodes", "GOids", envir=stam.env)
  }
}

#
#   End of File
#
###############################################################################
