################################################################################
#
#   04_stamCV.R
#
#   This piece of code performs a cross validation to determine the
#   most appropriate shrinkage level within a structured analysis of
#   microarray data (StAM). Thereby, PAM is fitted in each leaf node
#   prior to weigh all according to their performance. Results of inner
#   nodes are computed as weighted sums of their children.
#
#   written by Claudio Lottaz (COMPDIAG-MPI-MG), in September 2003
#
################################################################################
#
#   Class Definition and Constructor
#

setClass("stamCV", 
         representation(sample.labels  = "character",  # names of samples,
                        sample.classes = "character",  # classifications of samples
                        class.labels   = "character",  # names of classes [class]
                        prior          = "numeric",    # prior probability per class
                        beta           = "numeric",    # class weighting [class]
                        gamma          = "matrix",     # positive threshold [class]
                        full.pamfit    = "pamrtrained",# hold centroids etc. for all nodes
                        probs          = "array",      # cv. probs [samples,clases,nodes]
                        folds          = "list",       # buckets used in cross-valdation
                        results        = "data.frame", # err.rate/performance/redundancy per delta
                        node.results   = "list", # performance/redundancy/sens/spec per node
                        max.leafdev    = "numeric",    # max deviance in leaf nodes
                        deltas         = "numeric"),   # shrinkage factors to try
         contains = "stamNet")

setMethod("initialize", "stamCV", 
  function(.Object, exprs, classifications, beta, chip, root) {
    if (ncol(exprs) != length(classifications)) 
      stop("Exactly as many classifications as samples must be provides")

    .Object <- callNextMethod(.Object, chip, root, rownames(exprs))
    if (is.null(colnames(exprs))) .Object@sample.labels <- as.character(seq(ncol(exprs)))
    else .Object@sample.labels <- colnames(exprs)
    .Object@sample.classes <- as.character(classifications)

    class.table <- table(classifications)
    .Object@class.labels <- names(class.table)
    .Object@prior <- as.numeric(class.table/length(classifications))
    if (is.null(beta)) .Object@beta = .Object@prior
    else .Object@beta <- beta
    if (length(class.table) != length(.Object@beta)) 
      stop("'beta' must specify exactly one weight per classe in 'classifications'")
    if (sum(.Object@beta) != 1) stop("Weights in 'beta' must sum to 1")

    emptyfit <- list()
    class(emptyfit) <- "pamrtrained"
    .Object@full.pamfit <- emptyfit
    .Object@probs<-array(0, c(ncol(exprs), length(.Object@class.labels), length(.Object@nodes)),
                         dimnames=list(colnames(exprs),.Object@class.labels,names(.Object@nodes)))
    .Object@folds <- stam.balanced.folds(classifications)
    .Object@results <- data.frame("Root.error.rate"  = numeric(0),
                                  "Root.performance" = numeric(0),
                                  "Mean.redundancy"  = numeric(0),
                                  "Nodes"            = numeric(0),
                                  "Genes"            = numeric(0))
    .Object@node.results <- list()
    .Object@deltas <- numeric(0)
    .Object@max.leafdev <- 0
    return(.Object)
  }
)

################################################################################
#
#   Computing
#

reconstruct.pamfit <- function(fit, cv) {
  # Completes the restricted pamfit 'fit' using information
  # from the full fit in 'cv', the argument or the element of stam.env
  if (missing(cv)) {
    if (!exists("cv", stam.env)) stop("'cv' missing and no 'cv' exists in 'stam.env'")
    fit$y                 <- stam.env[["cv"]]@full.pamfit$y
    fit$centroids         <- stam.env[["cv"]]@full.pamfit$centroids[fit$gene.subset,]
    attr(fit$centroids,"scaled:scale") <- attr(stam.env[["cv"]]@full.pamfit$centroids,"scaled:scale")
    fit$centroid.overall  <- stam.env[["cv"]]@full.pamfit$centroid.overall[fit$gene.subset]
    fit$threshold.scale   <- stam.env[["cv"]]@full.pamfit$threshold.scale
    fit$se.scale          <- stam.env[["cv"]]@full.pamfit$se.scale
    fit$prior             <- stam.env[["cv"]]@full.pamfit$prior
    fit$sample.subset     <- stam.env[["cv"]]@full.pamfit$sample.subset
  } else { # cv is provided, so use it
    fit$y                 <- cv@full.pamfit$y
    fit$centroids         <- cv@full.pamfit$centroids[fit$gene.subset,]
    attr(fit$centroids,"scaled:scale") <- attr(cv@full.pamfit$centroids,"scaled:scale")
    fit$centroid.overall  <- cv@full.pamfit$centroid.overall[fit$gene.subset]
    fit$threshold.scale   <- cv@full.pamfit$threshold.scale
    fit$se.scale          <- cv@full.pamfit$se.scale
    fit$prior             <- cv@full.pamfit$prior
    fit$sample.subset     <- cv@full.pamfit$sample.subset
  }
  return(fit)
}

restrict.pamfit <- function(fit) {
  # remove unneeded elements from pam fit 'n'
  fit$y                 <- NULL
  fit$yhat              <- NULL
  fit$prob              <- NULL
  fit$centroids         <- NULL
  fit$centroid.overall  <- NULL
  fit$threshold.scale   <- NULL
  fit$se.scale          <- NULL
  fit$prior             <- NULL
  fit$sample.subset     <- NULL
  return(fit)
}

stam.fit.pam <- function(exprs, verbose=FALSE) {
  # fits one pam per leaf node, on the nodes genes only. 
  # a pamimagefile can be used as abbreviation
  # result is returned as side effect in cv in stam.env

  # fit pams
  cat(" - Training leaf nodes...\n")
  dat <- list(x = exprs, y = stam.env[["cv"]]@sample.classes)
  nleafs <- length(stam.env[["cv"]]@leafs)
  deviance <- rep(0, length=nleafs)
  cat("    . computing full fit...\n")
  sink(file="tmp.log", type="output")  
  stam.env[["cv"]]@full.pamfit <- pamr.train(dat); cat("\n")
  nodes <- stam.env[["cv"]]@nodes
  probs <- stam.env[["cv"]]@probs
  if (!verbose) {
    barlength <- getOption("width") - 9
    if (nleafs < barlength) barlength <- nleafs
    dotbucket <- nleafs / barlength
    sink()
    cat("    . computing leaf nodes...\n")
    cat("      0%")
    for (i in 1:(barlength - 6)) cat(".")
    cat("100%\n      ")
    sink(file="tmp.log", type="output")  
  }
  for (i in 1:nleafs) {
    n <- stam.env[["cv"]]@leafs[i]
    current.probes <- stam.env[["cv"]]@nodes[[n]]@probes
    if (length(current.probes) == 1) current.probes[2] <- current.probes[1]
    fit <- pamr.train(dat, gene.subset = current.probes); 
    cat("\n")
    cv <- pamr.cv(fit, dat, folds=stam.env[["cv"]]@folds)
    best.delta <- max(which(cv$error == min(cv$error)))
    nodes[[n]]@delta <- cv$threshold[best.delta]
    probs[,,n] <- fit$prob[,,best.delta]
    nodes[[n]]@pamfit <- restrict.pamfit(fit)
    if (verbose) {
      sink()
      cat("    . PAM best shrinkage for node ", nodes[[n]]@ID, " (", i, "/", nleafs,
          ") is ",nodes[[n]]@delta," yields error-rate ",cv$error[best.delta],"\n",sep="")
      sink(file="tmp.log", type="output")
    } else {
      if ((i %% dotbucket) < 1) {
        sink()
        cat("=")
        sink(file="tmp.log", type="output")
      }
    }
  }
  sink()
  file.remove("tmp.log")
  if (!verbose) cat("\n")
  stam.env[["cv"]]@probs <- probs
  stam.env[["cv"]]@nodes <- nodes
  invisible(NULL)
}

stam.weightnodes.fold <- function(cv.devs, delta, trainingset, testset) {
  # determine weights, probabilities and deviances for given fold and 
  # shrinkage. Results are returned as side-effect in cv in the stam.env
  probs <- stam.env[["cv"]]@probs
  sample.classes <- stam.env[["cv"]]@sample.classes[trainingset]
  for (n in stam.env[["cv"]]@inodes) {
    w <- stam.env[["cv"]]@max.leafdev - cv.devs[stam.env[["cv"]]@nodes[[n]]@children] - delta
    w[w < 0] <- 0
    s <- sum(w)
    if (s == 0) {
      probs[,,n] <- matrix(stam.env[["cv"]]@prior,nrow=dim(probs)[1], ncol=dim(probs)[2], byrow=TRUE)
    } else {
      w <- w / s;
      probs[,,n] <- .Call("sumprobs",probs,stam.env[["cv"]]@nodes[[n]]@children,w,PACKAGE="stam")
    }
    cv.devs[n] <- stam.deviance(probs[trainingset,,n], sample.classes, stam.env[["cv"]]@beta)
  }
  stam.env[["cv"]]@probs[testset,,] <- probs[testset,,]
  invisible(NULL)
}

stam.weightnodes <- function(cv.devs, delta) {
  # same as stam.weightnode.fold but doesn't care about training/testset
  # and stores weights in nodes instead of probs
  probs <- stam.env[["cv"]]@probs
  nodes <- stam.env[["cv"]]@nodes
  for (n in stam.env[["cv"]]@inodes) {
    w <- stam.env[["cv"]]@max.leafdev - cv.devs[stam.env[["cv"]]@nodes[[n]]@children] - delta
    w[w < 0] <- 0
    s <- sum(w)
    if (s == 0) {
      probs[,,n] <- matrix(stam.env[["cv"]]@prior, nrow=dim(probs)[1], ncol=dim(probs)[2], byrow=TRUE)
      nodes[[n]]@weights[] <- 0
    } else {
      w <- w / s;
      probs[,,n] <- .Call("sumprobs",probs,stam.env[["cv"]]@nodes[[n]]@children,w,PACKAGE="stam")
      nodes[[n]]@weights[] <- w
    }
    cv.devs[n] <- stam.deviance(probs[,,n], stam.env[["cv"]]@sample.classes, stam.env[["cv"]]@beta)
  }
  stam.env[["cv"]]@nodes <- nodes
  invisible(NULL)
}

stam.shrink <- function(expr.mat, deltas=NULL, ndeltas=10, results.per.node=FALSE,
                        collapse.scnodes = FALSE) {
  # Evaluated each candidate delta using cross-validation,
  # choosing weights according to inverse deviance.
  # Expects network and parameters in cv in stam.env

  nfolds <- length(stam.env[["cv"]]@folds)
  cv.devs <- matrix(0, nrow=length(stam.env[["cv"]]@nodes), ncol=nfolds+1)
  for (f in 1:nfolds) {
    testset <- stam.env[["cv"]]@folds[[f]]
    trainingset <- setdiff(1:ncol(expr.mat), testset)
    cv.devs[stam.env[["cv"]]@leafs,f] <- apply(stam.env[["cv"]]@probs[trainingset,,stam.env[["cv"]]@leafs],3,stam.deviance,
                                       stam.env[["cv"]]@sample.classes[trainingset],stam.env[["cv"]]@beta)
  }
  cv.devs[stam.env[["cv"]]@leafs, nfolds+1] <- apply(stam.env[["cv"]]@probs[,,stam.env[["cv"]]@leafs], 3, stam.deviance,
                                             stam.env[["cv"]]@sample.classes, stam.env[["cv"]]@beta)

  # determine reasonable deltas considering deviances in leafs 
  # if not provided by the user
  maxdev <- max(cv.devs[stam.env[["cv"]]@leafs, nfolds+1])
  stam.env[["cv"]]@max.leafdev <- maxdev
  if (!is.null(deltas)) stam.env[["cv"]]@deltas <- deltas
  else {
    devrange <- maxdev - min(cv.devs[stam.env[["cv"]]@leafs, nfolds+1])
    stam.env[["cv"]]@deltas <- seq(0, devrange, devrange/(ndeltas - 1))
  }

  stam.trydeltas(expr.mat, cv.devs, results.per.node, collapse.scnodes)
  return(stam.env[["cv"]])
}

stam.trydeltas <- function(expr.mat, cv.devs, results.per.node, collapse.scnodes) {
  results <- data.frame("Root.error.rate"  = numeric(0),
                        "Root.performance" = numeric(0),
                        "Mean.redundancy"  = numeric(0),
                        "Nodes"            = numeric(0),
                        "Genes"            = numeric(0))
  resultCtr <- 1
  node.results <- list()

  gamma <- numeric(0)
  nfolds <- length(stam.env[["cv"]]@folds)
  rootidx <- length(stam.env[["cv"]]@nodes)
  for (d in stam.env[["cv"]]@deltas) {
    if (length(stam.env[["cv"]]@deltas) == 1) 
      cat("    . fitting weights with shrinkage", d, "...\n")
    else cat("    . try shrinkage level", d, "...\n")
    for (f in 1:nfolds) {
      testset <- stam.env[["cv"]]@folds[[f]]
      trainingset <- setdiff(1:ncol(expr.mat), testset)
      stam.weightnodes.fold(cv.devs[, f], d, trainingset, testset)
    }
    dev <- stam.deviance(stam.env[["cv"]]@probs[,,rootidx], stam.env[["cv"]]@sample.classes, 
                         stam.env[["cv"]]@beta)

    # generate unique fit (not per fold)
    tmpprobs <- stam.env[["cv"]]@probs
    stam.weightnodes(cv.devs[, nfolds + 1], d)
    nz <- stam.nzgraph(stam.env[["cv"]])
    tmp <- stam.threshpreds(tmpprobs[,,nz$nodes, drop=FALSE], 
                               as.character(stam.env[["cv"]]@sample.classes))
    pred <- tmp$pred
    gamma <- rbind(gamma, tmp$gamma)
#    pred <- stam.maxpreds(tmpprobs[,,nz$nodes, drop=FALSE])[,1]
# approach considerung root node only
#    pred <- stam.env[["cv"]]@class.labels[apply(stam.env[["cv"]]@probs[,, rootidx], 1, 
#                                        function(v) which(v == max(v)))]
    err <- sum(pred != as.character(stam.env[["cv"]]@sample.classes)) / 
           length(stam.env[["cv"]]@sample.classes)

    remap <- rep(0, length(stam.env[["cv"]]@nodes))
    nodes <- list()
    for (n in 1:length(nz$nodes)) {
      remap[nz$nodes[n]] <- n
      nodes[[n]] <- stam.env[["cv"]]@nodes[[nz$nodes[n]]]
      names(nodes)[n] <- names(stam.env[["cv"]]@nodes)[nz$nodes[n]]
      if (class(nodes[[n]]) == "stamINode")
        nodes[[n]]@children[] <- remap[stam.env[["cv"]]@nodes[[nz$nodes[n]]]@children]
    }
    if (collapse.scnodes) { 
      tmp <- stam.collapse.scnodes(nodes)
      nodes <- tmp$nodes
      nz$nodes <- nz$nodes[tmp$nidxs]
    }
    if (length(nodes) < 2) {
      cat("      no nodes left after shrinkage.\n")
      break;
    }
    r <- .Call("redundancy", stam.env[["cv"]]@probs, nz$nodes, PACKAGE="stam")
    red <- r$avg
    if (results.per.node) {
      node.results$performances <- apply(stam.env[["cv"]]@probs[,,nz$nodes], 3, stam.deviance,
                                      stam.env[["cv"]]@sample.classes, stam.env[["cv"]]@beta)
      preds <- stam.preds(stam.env[["cv"]]@probs[,,nz$nodes], stam.env[["cv"]]@class.labels)
      err.aux <- function(v, cls) sum(v != cls)/length(cls)
      node.results$error.rate <- apply(preds, 2, err.aux, stam.env[["cv"]]@sample.classes)
      node.results$sensitivity <- stam.sensitivity(preds, stam.env[["cv"]]@sample.classes,
                                                   stam.env[["cv"]]@class.labels)
      node.results$specificity <- stam.specificity(preds, stam.env[["cv"]]@sample.classes,
                                                   stam.env[["cv"]]@class.labels)
      node.results$redundancy <- r$r    
    }
    cat("      yields error rate: ",err," (performance ", dev, ", redundancy ", red, ")\n",
        "      through ", length(nz$nodes), " nodes with ", length(nz$genes), " accessible genes\n",
        sep="")
    results[resultCtr,] <- c(err, dev, red, length(nz$nodes), length(nz$genes))
    resultCtr <- resultCtr + 1
  }
  if (length(stam.env[["cv"]]@deltas) == 1) { # return results in cv to fit
    stam.env[["cv"]]@nodes  <- nodes
    stam.env[["cv"]]@leafs  <- remap[intersect(stam.env[["cv"]]@leafs, nz$nodes)]
    stam.env[["cv"]]@inodes <- remap[intersect(stam.env[["cv"]]@inodes, nz$nodes)]
  }
  if (nrow(results) > 0) 
    rownames(results) <- round(stam.env[["cv"]]@deltas[1:nrow(results)], 3)
  stam.env[["cv"]]@gamma <- gamma
  stam.env[["cv"]]@results <- results;
  stam.env[["cv"]]@node.results <- node.results
  invisible(NULL)
}

stam.collapse.scnodes <- function(allnodes) {
  nodes <- list()
  nidxs <- NULL
  next.idx <- 1
  remap <- rep(0, length(allnodes))
  for (i in 1:length(allnodes)) {
    n <- allnodes[[i]]
    if (class(n) == "stamLeaf") {
      nodes[[next.idx]] <- n
      nidxs <- c(nidxs, i)
      remap[i] <- next.idx
      next.idx <- next.idx + 1
    } else {
      remapped.children <- remap[n@children]
      if (length(unique(remapped.children)) > 1) {
        nodes[[next.idx]] <- n
        nodes[[next.idx]]@children[n@children>0] <- remapped.children
        nidxs <- c(nidxs, i)
        remap[i] <- next.idx
        next.idx <- next.idx + 1
      } else { # remove an inner node
        ch <- unique(remapped.children)
        remap[i] <- ch
        nodes[[ch]]@replacedParents <- c(nodes[[ch]]@replacedParents, n@ID)
      }
    }
  }
  for (i in 1:max(remap)) {
    old.i <- min(which(remap == i))
    names(nodes)[i] <- names(allnodes)[old.i]
  }
  return(list(nodes=nodes, nidxs=nidxs))
}

################################################################################
#
#   Reports and HTML
#

setMethod("print", "stamCV",
  function(x, ...) {
    cat("stamCV on expression matrix holding ", length(x@sample.labels), " samples from ", 
        length(x@class.labels), " classes.\nStructure from: ", sep="")
    callNextMethod()
    cat("Class priors:")
    for (i in 1:length(x@class.labels)) 
      cat(" '", x@class.labels[i], "': ", x@prior[i], sep="")
    cat("\nClass weights:")
    for (i in 1:length(x@class.labels)) 
      cat(" '", x@class.labels[i], "': ", x@beta[i], sep="")
    cat("\nShrinkage candidates:", x@deltas, "\n")
    cat("\nResults from ", length(x@folds), "-fold cross validation:\n", sep="")
    print(round(x@results, 5))
  }
)

setMethod("writeHTML", "stamCV",
  function(x, aclass=NULL, delta=NULL, ...) {
    callNextMethod(x, nonodes=TRUE)
    prior <- as.numeric(table(x@sample.classes)/length(x@sample.classes))
    class.prior <- ""
    class.weights <- ""
    for (i in 1:length(x@class.labels)) {
      class.weights <- paste(class.weights," '",x@class.labels[i],"': ",round(x@beta[i],3),sep="")
      class.prior   <- paste(class.prior,  " '",x@class.labels[i], "':",round(prior[i], 3),sep="")
    }

    catHTML(HR(), NAME("cvinfo"), H2("Information on Cross Validation"), 
            P(TABLE(cellpadding="5", #cellspacing="5", border="5", 
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Number of samples:"), 
                               TD(length(x@sample.labels))),
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Number of classes:"), 
                               TD(length(x@class.labels))),
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Class weights:"), 
                               TD(class.weights)),
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Class prior:"), 
                               TD(class.prior)),
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Shrinkage candidates:"), 
                               TD(toString(round(x@deltas,3)))))))
    filenamestem <- sub(".html$", "", stam.env[["HTMLfile"]])
    plot(x, outfile=filenamestem, aclass=aclass, delta=delta)
    plot(x, outfile=filenamestem, aclass=aclass, delta=delta, ps=TRUE)
    if (regexpr("/$", filenamestem) == 1) filenamestem <- ""
    else {
      tmp <- strsplit(filenamestem, "/")[[1]]
      filenamestem <- tmp[length(tmp)]
    }
    catHTML(HR(), H2("Cross validated performance"),
            P(align="center", 
              IMG(src=paste(filenamestem, "_edr_plot.png", sep="")), "<BR>",
              HREF("Download postscript",
                   url=paste(filenamestem, "_edr_plot.ps", sep=""))),
            P(align="center", 
              IMG(src=paste(filenamestem, "_ng_plot.png", sep="")), "<BR>",
              HREF("Download postscript",
                 url=paste(filenamestem, "_ng_plot.ps", sep=""))),
            P(tableHTML(cellpadding="5", cell.valign="top", header.bgcolor="lightgreen", 
                        table.align="center", cell.align="center", round(x@results, 5))))
  }
)

################################################################################
#
#   Plots and Images
#

stam.add.delta <- function(delta, vertical=TRUE) {
  if (vertical) {
    abline(v=delta)
    mtext(expression(Delta[0]), side=3, line=0.25, at=delta, las=1)
  } else {
    abline(h=delta)
    mtext(expression(Delta[o]), side=4, line=0.25, at=delta, las=1)
  }
}

stam.edr.plot <- function(x, deltas, aclass=NULL, main=NULL, delta=NULL) {
  par(mfrow=c(3,1))
  if (is.null(aclass)) aclass <- x@class.labels[1]
  if (is.null(main)) main<-paste("Cross Validation of ",aclass,"-Prediction - Performance",sep="")
  par(mai=c(0.0,1.0,1.0,0.5))
  plot(deltas, 100*x@results$Root.error.rate, type="l",  main=main,
       xlab="", ylab="root error rate", axes=FALSE)
  axis(side=2, las=1); box()
  if(!is.null(delta)) stam.add.delta(delta)
  par(mai=c(0.5,1.0,0.5,0.5), las=1)
  plot(deltas, x@results$Root.performance, type="l",
       xlab="", ylab="root performance", axes=FALSE)
  axis(side=2, las=1); box()
  if(!is.null(delta)) stam.add.delta(delta)
  par(mai=c(1.0,1.0,0.0,0.5))
  plot(deltas, x@results$Mean.redundancy, type="l",
       xlab=expression(paste("shrinkage parameter ", Delta)), ylab="mean redundancy", las=1)
  if(!is.null(delta)) stam.add.delta(delta)
}

stam.ng.plot <- function(x, deltas, aclass=NULL, main=NULL, delta=NULL) {
  par(mfrow=c(2,1))
  if (is.null(aclass)) aclass <- x@class.labels[1]
  if (is.null(main)) main<-paste("Cross Validation of ", aclass, "-Prediction - Data Used",sep="")
  par(mai=c(0.25,1.0,1.0,0.5))
  plot(deltas, x@results$Nodes, type="l",  main=main,
       xlab="", ylab="number of GO-nodes", axes=FALSE)
  axis(side=2, las=1); box()
  if(!is.null(delta)) stam.add.delta(delta)
  par(mai=c(1.0,1.0,0.25,0.5))
  plot(deltas, x@results$Genes, type="l", las=1,
       xlab=expression(paste("shrinkage parameter ", Delta)), ylab="number of accessible genes")
  if(!is.null(delta)) stam.add.delta(delta)
}

plot.stamCV <- function(x, outfile=NULL, aclass=NULL, delta=NULL, main=NULL, which=0,
                        res=72, ps=FALSE, pointsize=16, ...) {
  deltas <- x@deltas[seq(nrow(x@results))]

  # error-rate/performance/redundancy plot
  if ((which == 1) || ((which == 0) && getOption("interactiveStam"))) {
    stam.opendev(outfile, "_edr_plot", ps, res, pointsize)
    stam.edr.plot(x, deltas, aclass, main, delta)
    if (!is.null(outfile)) dev.off()
  }
  if (is.null(outfile) && getOption("interactiveStam") && (which == 0)) 
    readline("<hit return to continue>")
  
  # nnodes/ngenes plot
  if ((which == 2) || ((which == 0) && getOption("interactiveStam"))) {
    pointsize = 0.66*pointsize      # compensate for mfrow(2,1) instead of mfrow(3,1)
    stam.opendev(outfile, "_ng_plot", ps, res, pointsize)
    stam.ng.plot(x, deltas, aclass, main, delta)
    if (!is.null(outfile)) dev.off()
  }

  invisible(NULL)
}

################################################################################
#
#   Odds and ends
#

stam.matchcv <- function(cv, exprs, cls, chip, root, beta, deltas, ndeltas) {
  # check whether cv is consistent with the other parameters
  if (class(cv) != "stamCV") {
    cat("   'cv' must be of class 'stamCV' instead of '", class(cv), "'\n", sep="")
    return(FALSE)
  }
  if (ncol(exprs) != length(cv@sample.labels)) {
    cat("   Number of sample precompiled pamimage doesn't match\n", sep="")
    return(FALSE)
  }
  if (sum(colnames(exprs) != cv@sample.labels) > 0) {
    cat("   Sample labels in precompiled pamimage don't match\n", sep="")
    return(FALSE)
  }
  for (i in 1:nrow(exprs)) {
    pid <- rownames(exprs)[i]
    if (exists(pid, cv@probes)) {
      if (i != get(pid, cv@probes)) {
        cat("   Order of rownames in expression matrix does not match pamimage\n", sep="")
        return(FALSE)
      }
    } else {
      cat("Expression matrix holds probes not present in pamimage, e.g.", pid, "\n", sep="")
      return(FALSE)
    }
  }
  if (sum(cls != cv@sample.classes) > 0) {
    cat("   Class labels do not match pamimage\n", sep="")
    return(FALSE)
  }
  if (chip != cv@chip) {
    cat("   Chip (", chip, ") is not the same in pamimage (", cv@chip, ")\n", sep="")
    return(FALSE)
  }
  if (root != cv@root) {
    cat("Root (", root, ") is not the same in pamimage (", cv@root, ")\n", sep="")
    return(FALSE)
  }
  if (!is.null(beta)) {
    if (sum(beta != cv@beta) > 0) {
      cat("   Overwriting class weights in pamimage\n", sep="")
      stam.env[["cv"]]@beta <- beta
    }
  }
  cat("   Precomputed pamimage looks consistent\n")
  return(TRUE)
}

#
#   Enf of File
#
################################################################################



