################################################################################
#
#   05_fit.R
#
#   This file contains code for fitting node weights for a particular
#   shrinkage leval delta in StAM (Structured Array Analysis). delta
#   is chosen automatically based on cross validation results if it
#   is not provided by the user. Only weights of nodes (edges) are
#   adapted, the results for leaf-nodes are taken from the cross-
#   validation. The user can either specify delta itself or the
#   weighting between deviance in the root node and the mean redun-
#   dancy over all nodes with weights. 
#
#   written by Claudio Lottaz (COMPDIAG-MPI-MG), in October 2003
#
################################################################################
#
#   Class Definition and Constructor
#

setClass("stamFit", 
         representation(class.labels     = "character",  # names of classes [class]
                        sample.classes   = "character",  # classifications of samples
                        prior            = "numeric",    # prior probability per class
                        full.pamfit      = "pamrtrained",# hold centroids etc f. all nodes
                        alpha            = "numeric",    # weight root dev. vs. mean red.
                        beta             = "numeric",    # class weighting [class]
                        gamma            = "numeric",    # prediction thesholds [classes]
                        collapse.scnodes = "logical",    # collapse single child nodes
                        delta          = "numeric",      # user suggested shrinkage choice
                        best.delta     = "numeric",      # chosen shrinkage level
                        default.delta    = "numeric",    # heuristic default for delta
                        scores         = "matrix",       # combination of red. and dev.
                        alpha.tab      = "matrix",       # compare alphas table
                        node.results   = "list"),        # cv-results per node
         contains = "stamNet")

setMethod("initialize", "stamFit", 
  function(.Object, cv, exprs, alpha, delta, max.nodes, collapse.scnodes) {
    # copy stamNet-data from cv
    .Object@chip    <- cv@chip
    .Object@root    <- cv@root
    .Object@chippkg <- cv@chippkg
    .Object@GOpkg   <- cv@GOpkg
    .Object@probes  <- cv@probes
    .Object@collapse.scnodes <- collapse.scnodes

    # copy stamCV-data from cv
    .Object@class.labels   <- cv@class.labels
    .Object@sample.classes <- cv@sample.classes
    .Object@prior          <- cv@prior
    .Object@full.pamfit    <- cv@full.pamfit
    .Object@beta           <- cv@beta

    # determine delta according to provided arguments
    .Object@alpha.tab <- matrix(0)
    .Object@scores <- matrix(0)
    .Object@default.delta <- numeric(0)
    if (is.null(alpha)) .Object@alpha <- numeric(0)
    else .Object@alpha <- alpha
    if ((length(alpha) > 0) && (!is.null(delta))) {
      .Object@default.delta <- delta
      delta <- NULL
    }
    if (is.null(delta)) {
      .Object@delta <- numeric(0)
      if (is.null(alpha)) { # determine delta from error-rate only
        tmp <- (cv@results$Root.error.rate == min(cv@results$Root.error.rate))
        d <- min(which(tmp & (cv@results$Nodes <= max.nodes)))
        if (!is.finite(d)) d <- max(which(tmp))
        .Object@best.delta <- round(cv@deltas[d], 4)
      } else { # determine delta from deviance and redundancy
        .Object@scores <- matrix(0, nrow=length(alpha), ncol=nrow(cv@results),
                                  dimnames=list(alpha, rownames(cv@results)))
        scaled.performance <- stam.scale(cv@results$Root.performance)
        scaled.redundancy <- stam.scale(cv@results$Mean.redundancy)
        for (a in 1:length(alpha))
          .Object@scores[a,] <- alpha[a]*scaled.performance + (1-alpha[a])*scaled.redundancy
        d <- apply(.Object@scores, 1, stam.findmin)
        .Object@alpha.tab <- cbind(round(cv@deltas[d], 4),
                                   round(cv@results$Root.error.rate[d], 4),
                                   round(cv@results$Root.performance[d], 4),
                                   round(cv@results$Mean.redundancy[d], 4),
                                   cv@results$Nodes[d], cv@results$Genes[d])
        rownames(.Object@alpha.tab) <- alpha
        colnames(.Object@alpha.tab) <- c("Best Delta", "Root Error Rate", "Performance", 
                                         "Redundancy", "#Nodes", "#Genes")
        if (length(unique(d)) == 1) {
          .Object@best.delta <- round(cv@deltas[d[1]], 4)
        } else {
          print(.Object@alpha.tab)
          if (length(.Object@default.delta) == 0) {
            node.ok <- min(which(as.numeric(.Object@alpha.tab[,5])<max.nodes), length(d))
            default.d <- d[max(ceiling(length(d)/2), node.ok)]
            .Object@default.delta <- round(cv@deltas[default.d], 4)
          }
          repeat {
            if (!getOption("interactiveStam")) ans = ""
            else {
              ans <- readline(paste("      Which delta should be used (default is ",
                              .Object@default.delta, ")? ", sep=""))
            }
            if (ans == "") .Object@best.delta <- .Object@default.delta
            else .Object@best.delta <- as.numeric(ans)
            if (is.na(.Object@best.delta)) 
              cat("Unable to convert '",ans,"' to numeric. Please, retry...\n",sep="")
            else break
          }
        }
      }
    } else { # just use the delta provided
      .Object@delta <- delta
      if (delta > cv@max.leafdev)
        stop("'delta' must not be larger than max.leafdev (", cv@max.leafdev, ")")
      .Object@best.delta <- delta
      .Object@default.delta <- delta
    }

    # determine weights and performance
    stam.env[["cv"]] <- cv
    stam.shrink(exprs, deltas=.Object@best.delta, results.per.node=TRUE, 
                collapse.scnodes = collapse.scnodes) # output:stam.env[["cv"]] in stam.env!
    .Object@nodes        <- stam.env[["cv"]]@nodes
    tmp <- abs(stam.env[["cv"]]@deltas - .Object@best.delta)
    .Object@gamma        <- stam.env[["cv"]]@gamma[which(tmp==min(tmp)),]
    .Object@node.results <- stam.env[["cv"]]@node.results
    .Object@leafs        <- stam.env[["cv"]]@leafs
    .Object@inodes       <- stam.env[["cv"]]@inodes
    rm("cv", envir=stam.env)
    return(.Object)
  }
)

stam.scale <- function(v) {
  minv <- min(v)
  maxv <- max(v)
  return((v - minv)/(maxv - minv))
}

stam.findmin <- function(v) {
  return(min(which(v == min(v))))
}


################################################################################
#
#   Reports and HTML
#

setMethod("print", "stamFit",
  function(x, ...) {
    cat("stamFit to predict from ",  length(x@class.labels), " classes.\nStructure from: ", sep="")
    callNextMethod()
    cat("Class priors:")
    for (i in 1:length(x@class.labels)) 
      cat(" '", x@class.labels[i], "': ", round(x@prior[i], 4), sep="")
    cat("\nClass weights:")
    for (i in 1:length(x@class.labels)) 
      cat(" '", x@class.labels[i], "': ", round(x@beta[i], 4), sep="")
    cat("\nShrinkage parameter:", x@best.delta, " chosen by ")
    if (length(x@delta) != 0) cat("the user")
    else {
      if (length(x@alpha) == 0) cat("optimizing the error rate")
      else 
        cat("considering weighted scores between root performance ",
            "and mean redundancy\nCandidates for performance weights: ", 
             toString(x@alpha), sep="")
    }
    cat("\nLeft ", length(x@nodes), " nodes with weights\n")
    res <- round(cbind(x@node.results$performance, x@node.results$redundancy,
                       x@node.results$sensitivity, x@node.results$specificity), 4)
    colnames(res) <- c("Performance", "Redundancy",
                       paste(x@class.labels, "(sens.)"),
                       paste(x@class.labels, "(spec.)"))
    print(res)
  }
)

setMethod("writeHTML", "stamFit",
  function(x, exprs, aclass=NULL, nonodes=FALSE, write.form=FALSE, ...) {
    class.prior <- ""
    class.weights <- ""
    class.gamma <- ""
    for (i in 1:length(x@class.labels)) {
      class.weights <- paste(class.weights," '",x@class.labels[i],"': ",round(x@beta[i],3),sep="")
      class.prior   <- paste(class.prior,  " '",x@class.labels[i], "':",round(x@prior[i],3),sep="")
      class.gamma   <- paste(class.gamma,  " '",x@class.labels[i], "':",round(x@gamma[i],3),sep="")
    }
    if (length(x@delta) != 0) choice.method <- "the user"
    else {
      if (length(x@alpha) == 0) choice.method <- "optimizing the root's error rate"
      else
        if (length(x@alpha) == 1) 
          choice.method <- paste("optimizing a weighted score between root performance (", 
                                 x@alpha,") and mean redundancy (", 1-x@alpha,")", sep="")
        else {
          if (x@best.delta == x@default.delta) 
            choice.method <- "accepting the system default (after consulting weighting alternatives)"
          else choice.method <- "the user (after consulting weighting alternatives)"
        }
    }

    # compute accessible genes
    genes <- NULL
    for (i in 1:length(x@nodes)) {
      if (class(x@nodes[[i]]) != "stamLeaf") next
      genes <- unique(c(genes, x@nodes[[i]]@probes))
    }

    catHTML(HR(), NAME("fitinfo"), H2("Information on Model Fit"))

    filenamestem <- sub(".html$", "", stam.env[["HTMLfile"]])
    if (regexpr("/$", filenamestem) == 1) relfilenamestem <- ""
    else {
      tmp <- strsplit(filenamestem, "/")[[1]]
      relfilenamestem <- tmp[length(tmp)]
    }
    plot(x, outfile=filenamestem, aclass=aclass)
    plot(x, outfile=filenamestem, aclass=aclass, ps=TRUE)
    map <- stam.graph.plot(x, outfile=filenamestem)
    if (length(x@scores) > 1) {
      catHTML(HR(), H2("Performance vs. redundancy"),
              P(align="center",
                IMG(src=paste(relfilenamestem, "_ad_plot.png", sep="")),"<BR>",
                HREF("Download postscript", 
                     url=paste(relfilenamestem, "_ad_plot.ps", sep=""))))
    }
    if (ncol(x@alpha.tab) == 6)
      catHTML(P(tableHTML(cellpadding="5", cell.valign="top",header.bgcolor="lightgreen",
                          table.align="center",cell.align="center",round(x@alpha.tab,5))))
    if (write.form) 
      catHTML(stam.refit.form(filenamestem, aclass, x@best.delta, x@collapse.scnodes))

    catHTML(HR(),NAME("refit"),H2("Fit for <FONT face='Symbol'>D</FONT> = ",x@best.delta),
            P(TABLE(cellpadding="5", #cellspacing="5", border="5", 
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Number of classes:"), 
                               TD(length(x@class.labels))),
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Class weights:"), 
                               TD(class.weights)),
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Class prior:"), 
                               TD(class.prior)),
#              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Prediction thresholds:"), 
#                               TD(class.gamma)),
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Chosen shrinkage level:"), 
                               TD(round(x@best.delta, 3), "<BR>(by ", choice.method, ")", sep="")),
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Number of nodes left"), 
                               TD(length(x@nodes))),
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Number of accessible genes"), 
                               TD(length(genes))))),
            H3("Shrunken prediction graph"),
            MAP(name='graph_map', map),
            P("Hover over a node to see its description, click on it to view its ",
              "detailed results."),
            P(align="center",
              IMG(usemap='#graph_map',src=paste(relfilenamestem,"_graph_plot.png",sep="")),
              "<BR>", HREF("Download postscript",
                           url=paste(relfilenamestem, "_graph_plot.ps", sep=""))))
    catHTML(H3("Nodewise evaluation"),
            P(align="center",
              IMG(src=paste(relfilenamestem, "_fitnodes_plot.png",sep="")),"<BR>",
              HREF("Download postscript",
                   url=paste(relfilenamestem, "_fitnodes_plot.ps", sep=""))))

    if (nonodes) return(invisible(NULL))
    catHTML(HR(), NAME("nodesinfo"), H2("Details and Performance of GO-nodes"))
    stam.env[["allnodes"]] <- x@nodes
    stam.env[["GOids"]] <- new.env(hash=TRUE)
    for (i in 1:length(stam.env[["allnodes"]])) 
      assign(stam.env[["allnodes"]][[i]]@ID, i, stam.env[["GOids"]])
    if (!exists("cv", envir=stam.env)) stam.env[["cv"]] <- x
    if (!missing(exprs)) stam.env[["exprs"]] <- exprs
    for (i in length(x@nodes):1) {
      writeHTML(x@nodes[[i]], sepfiles=FALSE)
      stam.nodePerformanceHTML(x, i, "Performance on training data")
    }
    rm("cv", "allnodes", "GOids", envir=stam.env)
    if (!missing(exprs)) rm("exprs", envir=stam.env)
    invisible(NULL)
  }
)

stam.refit.form <- function(filenamestem, aclass, delta, collapse.scnodes=FALSE) {
  if (stam.cgi.url == "") {
    warning("No base URL for StAM's CGI scripts specified, no HTML forms written.\n",
            "Make sure to initialize the StAM server feature properly by calling \n",
            "stam.serve().\n")
    return("")
  }
  return(FORM(name="refitform", action=paste(stam.cgi.url, "refit_model.pl", sep="/"),
              HIDDEN(name="filenamestem", value=filenamestem),
              HIDDEN(name="aclass", value=aclass),
              UL(LI("Choose shrinkage level <FONT face='Symbol'>D</FONT>:", 
                    TEXTFIELD(name = "delta", value = delta, size = 5),
                    SUBMIT(name="Refit Model",value="Refit Model")),
                 LI("Collapse single-child nodes", 
                    CHECKBOX(name="collapse_scnodes", checked=collapse.scnodes)))))
}

stam.nodePerformanceHTML <- function(x, i, title) {
  res <- cbind(x@node.results$sensitivity[i,],x@node.results$specificity[i,])
  colnames(res) <- c("Sensitivity", "Specificity")
  rownames(res) <- x@class.labels
  catHTML(H3(title),
          P(UL(LI("Error rate:",   round(x@node.results$error.rate[i], 5)),
               LI("Performance:",   round(x@node.results$performance[i], 5)),
               LI("Redundancy:", round(x@node.results$redundancy[i], 5)))),
          P(tableHTML(cellpadding="5", cell.valign="top", header.bgcolor="lightgreen",
                      cell.align="center", round(res, 5))))
}

################################################################################
#
#   Plots and Images
#

stam.ad.plot <- function(x) {
  par(mfrow=c(1,2))
  deltas <- as.numeric(colnames(x@scores))
  best.deltas <- deltas[apply(x@scores, 1, stam.findmin)]
  
  # score vs. deltas
  plot(deltas, x@scores[1,], type="l", las=1,
       main="Weighted Scores vs. Shrinkage", ylab="weighted score",
       xlab=expression(paste("shrinkage parameter ", Delta)))
  points(best.deltas[1], min(x@scores[1,]), pch=19)
  if (nrow(x@scores) > 1) {
    for (a in 2:nrow(x@scores)) {
      lines(deltas, x@scores[a,])
      points(best.deltas[a], min(x@scores[a,]), pch=19)
    }
  }
  stam.add.delta(x@best.delta)

  # alpha vs. best delta
  plot(as.numeric(x@alpha), best.deltas, type="l", las=1,
       main="Performance Weight vs. Best Shrinkage", xlab="performance weight",
       ylab=expression(paste("best shrinkage parameter ", Delta)))
  stam.add.delta(x@best.delta, vertical=FALSE)
  invisible(NULL)
}

stam.nodes.plot <- function(x, aclass=NULL, minspec=NULL, minsens=NULL, maxsens=NULL) {
  par(mfrow=c(1,2))
  if (is.null(aclass)) aclass <- x@class.labels[1]
  plot(x@node.results$performance, x@node.results$redundancy, pch=19, cex=0.5, las=1,
       main="Performance vs. Redundancy", xlab="performance", ylab="redundancy")
  aclass <- sub(" \\(spec.\\)", "", colnames(x@node.results$specificity)) == aclass
  plot(x@node.results$sensitivity[,aclass], x@node.results$specificity[,aclass], pch=19, cex=0.5,
       main="Sensitivity vs. Specificity", xlab="sensitivity", ylab="specificity", las=1)
  if (!is.null(minspec)) abline(h=minspec)
  if (!is.null(minsens)) abline(v=minsens)
  if (!is.null(maxsens)) abline(v=maxsens)
  invisible(NULL)
}

stam.graph.plot <- function(x, outfile=NULL, pointsize=10, width=9, height=6) {
  # check needs 
  if (is.null(getOption("stam.dotcmd"))) {
    if (.Platform$OS.type == "unix") {
      if (system("which dot > /dev/null", ignore.stderr=TRUE) != 0) return()
      options(stam.dotcmd = system("which dot", intern=TRUE, ignore.stderr=TRUE))
    } else { # under windows
      if (system("dot --help") >= 0) options(stam.dotcmd = "dot")
      else return()
    }
  }
  if (is.null(outfile)) {
    warning("No direct graph output possible, generating png/ps.")
    outfile <- "graph_plot"
  } else outfile <- paste(outfile, "_graph_plot", sep="")

  # write dot-file - header
  con <- file(paste(outfile, ".dot", sep=""), open="w")
  cat('digraph {\ncenter=TRUE;\nrankdir=LR;\nsize="', width, ',', height, '";\n', 
      'node [style=filled,color=yellow,fontsize="',pointsize,'"];\n',
      'edge [dir=back];\n',
      sep = "", file = con)

  # write dot-file nodes
  for (n in 1:length(x@nodes)) {
    cnode <- x@nodes[[n]]
    id <- ifelse(cnode@supNode, paste(cnode@ID, "a", sep=""), cnode@ID)
    cat('"',id,'" [URL="', sub(":","",id), '.html",tooltip="', getGOterm(cnode),
        ifelse(class(cnode)=="stamLeaf", " (annotations)", ""), '"];\n', sep="",file=con)
  }

  # write dot-file - edges
  for (n in 1:length(x@nodes)) {
    from.node <- x@nodes[[n]]
    if (class(from.node) == "stamLeaf") next
    from.id <- ifelse(from.node@supNode, paste(from.node@ID, "a", sep=""), from.node@ID)
    for (ch in 1:length(from.node@children)) {
      if (from.node@weights[ch] == 0) next
      to.node <- x@nodes[[from.node@children[ch]]]
      to.id <- ifelse(to.node@supNode, paste(to.node@ID, "a", sep=""), to.node@ID)
      w <- round(from.node@weights[ch], 2)
      cat('"', from.id, '" -> "', to.id, '" [label="',w,'"];\n', sep="", file=con)
    }
  }

  # write dot-file - footer
  cat("}\n", file=con)
  close(con)

  # call dot to generate png, ps an image map
  system(paste(getOption("stam.dotcmd"), " -Tpng -o ", outfile, ".png ", 
               outfile, ".dot",sep=""))
  system(paste(getOption("stam.dotcmd"), " -Tps -o ", outfile, ".ps ", 
               outfile, ".dot", sep=""))
  map <- system(paste(getOption("stam.dotcmd"), " -Tcmap ", outfile, ".dot", sep=""), 
                intern=TRUE)

  # return map-contents
  map <- gsub(' alt="GO:[0-9]*a?" ', ' target="NodeInfo" ', map)
  invisible(map)
}


plot.stamFit <- function(x, outfile=NULL, aclass=NULL, main=NULL, which=0,
                        res=72, ps=FALSE, pointsize=12, ...) {

  # alpha/delta plots (if x@scores holds data)
  if ((which == 1) || ((which == 0) && getOption("interactiveStam"))) {
    if (length(x@scores) > 1) {
      stam.opendev(outfile, "_ad_plot", ps, res, pointsize)
      stam.ad.plot(x)
      if (!is.null(outfile)) dev.off()
      if (is.null(outfile) && getOption("interactiveStam") && (which == 0)) 
        readline("<hit return to continue>")
    }
  }
  
  # nodewise scatter-plots
  if ((which == 2) || ((which == 0) && getOption("interactiveStam"))) {
    stam.opendev(outfile, "_fitnodes_plot", ps, res, pointsize)
    stam.nodes.plot(x, aclass)
    if (!is.null(outfile)) dev.off()
  }
  
  invisible(NULL)
}


#
#   Enf of File
#
################################################################################



