################################################################################
#
#   06_stamPredict.R
#
#   This file contains the code needed to predict the casses of gene
#   expression profiles according to StAM (Structured Analysis of
#   Microarrays).
#
#   written by Claudio Lottaz (MPI-MG) in October 2003
#
################################################################################
#
#   Class Definition and Constructor
#


setClass("stamPrediction",
         representation(chip         = "character",    # chip taken from fit
                        nodes        = "list",         # stamINodes and stamLeafs
                        class.labels = "character",  # names of classes [class]
                        best.delta   = "numeric",    # chosen shrinkage parameter on edge weights
                        cls          = "character",    # known classifications
                        probs        = "array",        # pred. probs [samples, classes, nodes]
                        predicts     = "character",    # MAX predictions [samples]
                        testset      = "numeric",      # indices to test samples
                        node.results = "list"))        # test results per node

setMethod("initialize", "stamPrediction",
  function(.Object, fit, expr.mat, cls, testset) {
    if (!is.matrix(expr.mat)) expr.mat <- matrix(expr.mat, nrow=length(expr.mat), ncol=1)

    nnodes <- length(fit@nodes)
    .Object@chip         <- fit@chip
    .Object@class.labels <- fit@class.labels
    .Object@best.delta   <- fit@best.delta
    .Object@nodes        <- fit@nodes
    .Object@testset      <- testset

    nodenames <- names(fit@nodes)
    for (i in 1:length(nodenames)) {
      if (fit@nodes[[i]]@supNode) 
        nodenames[i] <- paste(nodenames[i], "a", sep="")
    }
    .Object@probs <- array(0, c(ncol(expr.mat), length(fit@class.labels), nnodes),
                           dimnames=list(colnames(expr.mat),fit@class.labels,nodenames))
    .Object@node.results <- list()
    predicts <- matrix("", nrow=ncol(expr.mat), ncol=nnodes, 
                       dimnames = list(colnames(expr.mat), names(fit@nodes)))
    .Object@node.results$predicts <- predicts

    # compute predicted class probabilities    
    predict.aux <- function(v, cl) cl[which(v == max(v))[1]]
    for (i in 1:nnodes) {
      n <- fit@nodes[[i]]
      if (class(n) == "stamLeaf") {
        pamfit <- reconstruct.pamfit(n@pamfit, fit)
        .Object@probs[,,i] <- pamr.predict(pamfit, expr.mat[n@pamfit$gene.subset,],
                                            n@delta, type="posterior")
      } else { # for inner nodes
        nchildren <- length(n@children)
        for (c in 1:nchildren) {
          if (n@children[c] > 0) 
            .Object@probs[,,i] <- .Object@probs[,,i] + 
              n@weights[c] * .Object@probs[,,n@children[c]]
        }
      }
      .Object@node.results$predicts[,i] <- 
                                   apply(matrix(.Object@probs[,,i], nrow=ncol(expr.mat)),
                                         1, predict.aux, fit@class.labels)
    }

    .Object@predicts <- stam.threshpreds(.Object@probs, gamma=fit@gamma)$pred
    names(.Object@predicts) <- colnames(expr.mat)

    # determine node.results
    if (is.null(cls)) { .Object@cls <- character(0) }
    else {
      .Object@cls <- cls
      if (ncol(expr.mat) == length(cls)) {
        evalset <- testset
        evalcls <- cls[testset]
      } else {
        if (ncol(expr.mat) - length(testset) == length(cls)) {
          evalset <- -testset # all but the testset
          evalcls <- cls
        } else {
          warning("'cls' must contain one entry for each sample or training sample")
          return(.Object)
        }
      }
      .Object@node.results$performance <- apply(.Object@probs[evalset,,,drop=FALSE], 3,
                                              stam.deviance, evalcls, fit@beta)
      .Object@node.results$sensitivity <- 
                         stam.sensitivity(.Object@node.results$predicts[evalset,,drop=FALSE],
                                          evalcls, fit@class.labels)
      .Object@node.results$specificity <- 
                         stam.specificity(.Object@node.results$predicts[evalset,,drop=FALSE],
                                          evalcls, fit@class.labels)
      if (length(evalset) == 1) {
        .Object@node.results$error.rate <- 
                     as.numeric(.Object@node.results$predicts[evalset,] != cls[evalset])
        .Object@node.results$redundancy <- rep(NA, length(fit@nodes))
      } else {
        err.aux <- function(v, cls) sum(v != cls)/length(cls)
        .Object@node.results$error.rate <- apply(.Object@node.results$predicts[evalset,],
                                                 2, err.aux, evalcls)
        r <- .Call("redundancy", .Object@probs[evalset,,], 1:dim(.Object@probs)[3], 
                   PACKAGE="stam")
        .Object@node.results$redundancy <- r$r
      }
    }
    return(.Object)
  }
)
         

################################################################################
#
#   Reports and HTML
#

setMethod("print", "stamPrediction",
  function(x, ...) {
    cat("stamPrediction on ",  nrow(x@node.results$predicts), " samples.\n", sep="")
    cat("Root predictor says: \n")
    print(x@node.results$predicts[,length(x@nodes)])
#    cat("MAX predictor says: \n") 
#    print(x@predicts)
  }
)

setMethod("writeHTML", "stamPrediction",
  function(x, aclass=NULL, minspec=NULL, minsens=0.1, maxsens=1.0, 
           nonodes=FALSE, write.form=FALSE, ...) {
    if (is.null(aclass)) aclass <- x@class.labels[1]
    predicts <- rep("", length(x@predicts))
    tmp <- strsplit(x@predicts, "/")
    for (i in 1:length(predicts))
      predicts[i] <- ifelse(aclass %in% tmp[[i]], aclass, "control")
    testset.entry <- ""
    if (length(x@cls) != length(predicts)) t <- ""
    else {
      t <- table(x@cls)
      t <- paste("(", toString(paste(t, "from", names(t))), ")", sep="")
      cls <- ifelse(x@cls == aclass, aclass, "control")
      max.err <- round(sum(cls[x@testset] != predicts[x@testset])/length(x@testset), 4)
      if (length(x@testset) < length(x@cls)) {
        tmp <- table(x@cls[x@testset])
        tmp <- paste("(", toString(paste(tmp, "from", names(tmp))), ")", sep="")
        testset.entry <- TR(valign="top",
                            TH(align="right",bgcolor="lightgreen","Test samples:"),
                            TD(length(x@testset), tmp))
      }
    }
    catHTML(HR(), NAME("predinfo"), H2("Information on Predictions"), 
            P(TABLE(cellpadding="5", #cellspacing="5", border="5", 
              TR(valign="top",TH(align="right",bgcolor="lightgreen","Number of samples:"),
                               TD(length(x@predicts), t)), testset.entry,
              TR(valign="top", TH(align="right", bgcolor="lightgreen", "Classes:"), 
                               TD(toString(dimnames(x@probs)[[2]])))#,
#              ifelse(length(x@cls) != length(predicts),"",
#                     TR(valign="top", 
#                       TH(align="right", bgcolor="lightgreen", 
#                          "MAX-predictor's error-rate:"), TD(max.err)))
              )))

    # MAX-prediction results
#    if (length(x@cls) == length(predicts)) {
#      clabels <- dimnames(x@probs)[[2]]
#      sens <- stam.sensitivity(x@predicts[x@testset], x@cls[x@testset], clabels)
#      spec <- stam.specificity(x@predicts[x@testset], x@cls[x@testset], clabels)
#      perf <- round(cbind(t(sens), t(spec)), 4)
#      colnames(perf) <- c("Sensitivity", "Specificity")
#      rownames(perf) <- clabels
#      tmpHTML <- tableHTML(cellpadding="5", cell.valign="top", 
#                           header.bgcolor="lightgreen",cell.align="center",
#                           round(perf[aclass,,drop=FALSE], 4))
#      catHTML(H3("Performance on testdata"), UL(tmpHTML))      
#    }
#    if (length(x@cls) != length(predicts)) 
#      predict.tab <- matrix(predicts, ncol=1,
#                            dimnames = list(names(x@predicts), "Predicted Class"))
#    else {
#      predict.tab <- cbind(predicts, x@cls)
#      rownames(predict.tab) <- names(x@predicts)
#      colnames(predict.tab) <- c("Predicted Class", "Labelled Class")
#      wrongs <- cls != predicts
#      predict.tab[wrongs,1] <- paste("<FONT COLOR='red'>", predicts[wrongs], "</FONT>")
#    }
#    predict.tab <- predict.tab[x@testset,]
#    tmpHTML <- tableHTML(cellpadding="5", cell.valign="top", header.bgcolor="lightgreen",
#                         cell.align="left", predict.tab)
#    tmpHTML <- gsub("GO:([0-9]+a?)", 
#                    "<A HREF='GO\\1.html' TARGET='NodeInfo'>GO:\\1</A>", 
#                    tmpHTML)
#    catHTML(H3("Prediction results per sample"), UL(tmpHTML))

    # plots
    filenamestem <- sub(".html$", "", stam.env[["HTMLfile"]])
    if (regexpr("/$", filenamestem) == 1) relfilenamestem <- ""
    else {
      tmp <- strsplit(filenamestem, "/")[[1]]
      relfilenamestem <- tmp[length(tmp)]
    }
    map <- image(x, outfile=filenamestem, aclass=aclass,
                 minspec=minspec, minsens=minsens, maxsens=maxsens)
    minspec <- image(x, outfile=filenamestem, aclass=aclass, 
                     minspec=minspec, minsens=minsens, maxsens=maxsens, ps=TRUE)
    catHTML(HR(), H2("Illustration of Performance on testdata"))
    if (length(x@cls) == length(predicts)) {
      plot(x, outfile=filenamestem, aclass=aclass,
           minspec=minspec, minsens=minsens, maxsens=maxsens)
      plot(x, outfile=filenamestem, aclass=aclass, 
           minspec=minspec, minsens=minsens, maxsens=maxsens, ps=TRUE)
      catHTML(H3("Nodewise evaluation"),
              P(align="center",
                IMG(src=paste(relfilenamestem, "_prednodes_plot.png", sep="")),
                "<BR>", HREF("Download postscript",
                             url=paste(relfilenamestem, "_prednodes_plot.ps", 
                                       sep=""))))
    }
    catHTML(NAME("pred_img"), H3("Prediction image"))
    if (file.exists(paste(filenamestem, "_pred_img.png", sep=""))) {
      catHTML(P("Hover over a node description to see its ID, click on it ", 
                "to see detailed results."),
              P(align="center",
               IMG(usemap="#pred_map",src=paste(relfilenamestem,"_pred_img.png",sep="")),
              "<BR>", HREF("Download postscript",
                           url=paste(relfilenamestem, "_pred_img.ps", sep=""))),
              "\n", MAP(name="pred_map", map))
    } else catHTML(P("No nodes selected."))
    if (write.form) {
      pred.img.form <- stam.pred.img.form(filenamestem, aclass, 
                                          minspec, minsens, maxsens)
      catHTML(pred.img.form, "\n")
    }
    if (nonodes) return(invisible(NULL))
    catHTML(HR(), NAME("nodesinfo"), H2("Details in GO-nodes"))
    stam.env[["allnodes"]] <- x@nodes
    stam.env[["GOids"]] <- new.env(hash=TRUE)
    for (i in 1:length(stam.env[["allnodes"]])) 
      assign(stam.env[["allnodes"]][[i]]@ID, i, stam.env[["GOids"]])
    for (i in length(x@nodes):1) {
      writeHTML(x@nodes[[i]], sepfiles=FALSE)
      if (length(x@cls) == length(predicts)) 
        stam.nodePerformanceHTML(x, i, "Performance on test data")
      stam.nodePredictHTML(x, i)
    }
    rm("allnodes", "GOids", envir=stam.env)
    invisible(NULL)
  }
)

stam.nodePredictHTML <- function(x, i) {
  if (length(x@cls) == 0) cls <- NULL
  else {
    if (length(x@cls) == length(x@predicts)) cls <- x@cls
    else {
      if (length(x@testset) + length(x@cls) != length(x@predicts)) {
        warning("Number of predicts seems inconsistent:", length(x@predicts), 
                " predictes, ", length(x@cls), " training and ", length(x@testset), 
                " test samples")
        return()
      }
      cls <- rep("", length(x@predicts))
      cls[-x@testset] <- x@cls
    }
  }

  res <- cbind(round(x@probs[,,i], 5), x@node.results$predicts[,i])
  if (!is.null(cls)) {
    res <- cbind(cls, res)
    colnames(res)[1] <- "Class"
    wrongs <- cls != x@node.results$predicts[,i]
    wrongs[cls == ""] <- FALSE
    rights <- cls == x@node.results$predicts[,i]
    rights[cls == ""] <- FALSE
     res[wrongs,ncol(res)] <- paste("<FONT COLOR='red'>", res[wrongs,ncol(res)], "</FONT>")
     res[rights,ncol(res)] <- paste("<FONT COLOR='green'>",res[rights,ncol(res)],"</FONT>")
  }
  colnames(res)[ncol(res)] <- "Predictionn"
  if (length(x@testset) > 0) 
    rownames(res)[x@testset] <- paste(rownames(res)[x@testset], "(testset)")
  catHTML(H3("Predictions"),
          P(tableHTML(cellpadding="5", cell.valign="top", header.bgcolor="lightgreen",
                      cell.align="center", res)))
}

stam.pred.img.form <- function(filenamestem, aclass, 
                               minspec, minsens, maxsens, bin.thresh="-",
                               f.red=0.0, f.green=1.0, f.blue=0.0, 
                               t.red=0.0, t.green=0.0, t.blue=0.0,
                               show.graph=TRUE, lab.len=40, sample.labels=FALSE) {
  if (stam.cgi.url == "") { 
    warning("No base URL for StAM's CGI scripts specified, no HTML forms written.\n",
            "Make sure to initialize the StAM server feature properly by calling \n",
            "stam.serve().\n")
    return("")
  }
  return(FORM(name="pred_img", action=paste(stam.cgi.url, "redraw_pred_img.pl", sep="/"),
              HIDDEN(name="filenamestem", value=filenamestem),
              HIDDEN(name="aclass", value=aclass),
              TABLE(align="center", cellspacing=20,
                    TR(TH("Node Selection"), TH("Colors"), TH("Options")),
                    TR(valign="top",
                       TD(align="right",
                          P(align="right","Min. specificity:", 
                            TEXTFIELD(name="minspec", value=minspec, size=5)),
                         P(align="right","Min. sensitivity:", 
                            TEXTFIELD(name="minsens", value=minsens, size=5)),
                          P(align="right","Max. sensitivity:", 
                            TEXTFIELD(name="maxsens", value=maxsens, size=5)),
                          P(align="right","Binery threshold", 
                            TEXTFIELD(name="bin_thresh", value=bin.thresh, size=3))),
                       TD(TABLE(TR(TH(P(align="right","from:")), TH(P(align="right","to:"))),
                                TR(valign="top", 
                                   TD(align="right", 
                                      P(align="right","red:",
                                        TEXTFIELD(name="f_red",value=f.red,size=3)),
                                      P(align="right","green:", 
                                        TEXTFIELD(name="f_green",value=f.green,size=3)),
                                      P(align="right","blue:", 
                                        TEXTFIELD(name="f_blue",value=f.blue,size=3))),
                                   TD(align="right", 
                                      P(align="right","red:",
                                        TEXTFIELD(name="t_red",value=t.red,size=3)),
                                      P(align="right","green:",
                                        TEXTFIELD(name="t_green",value=t.green,size=3)),
                                      P(align="right","blue:",
                                        TEXTFIELD(name="t_blue",value=t.blue,size=3)))))),
                       TD(align="right",
                          P(align="right",CHECKBOX(name="show_graph", checked=show.graph),
                             "Show GO-structure"),
                          P(align="right",CHECKBOX(name="sample_labels",checked=sample.labels),
                            "Show sample labels"),
                          P(align="right","Label length",TEXTFIELD(name="lab_len",value=lab.len,size=3)), 
                          P(align="right",SUBMIT(name="Update Image",value="Update Image")))))))
}

################################################################################
#
#   Plots and Images
#

stam.rgb.colors <- function(n=12, m=1,r0=1, g0=0, b0=0, r1=0, g1=1, b1=0) {
  tmp <- c(rep(0,n), seq(0, 1, 1/(n-1)))^(1/m)
  return(rgb(r0*rev(tmp)+r1*tmp, g0*rev(tmp)+g1*tmp, b0*rev(tmp)+b1*tmp))
}

collect.lines <- function(current.node, last.parent, depth, select, nodes) {
  l <- NULL
  cnode <- nodes[[current.node]]
  depth <- depth + length(cnode@replacedParents)
  if (current.node %in% select) {
    stam.env[["root.dist"]][select==current.node] <- max(stam.env[["root.dist"]][select==current.node], depth)
    last.parent <- current.node
  }
  if (class(cnode) == "stamLeaf") return(l)
  for (ch in unique(cnode@children)) {
    if (ch == 0) next    
    l <- rbind(l, collect.lines(ch, last.parent, depth+1, select, nodes))
    if (is.null(last.parent)) next   
    if (ch %in% select) l <- rbind(l, c(which(last.parent==select), which(ch==select)))
  }
  return(l)
}


image.stamPrediction <- function(x, aclass=NULL, main=NULL, xlab=NULL, show.graph=TRUE,
                                 max.label.length=40, sample.labels=FALSE, full.names=TRUE, 
                                 outfile=NULL, ps=FALSE, res=72, pointsize=10, width=11,
                                 minspec=NULL, minsens=0.1, maxsens=1.0, invert=FALSE,
                                 col=stam.rgb.colors(r0=0), bin.thresh=NULL, ...) {

  # determine probabilities to be shown and left labels
  if (is.null(aclass)) aclass <- dimnames(x@probs)[[2]][1]
  if(length(x@cls) == 0) {
    probs <- x@probs[,aclass,]
    ids <- colnames(probs)
    select <- rep(TRUE, ncol(probs))
    if (full.names) left.labels <- NULL
    else left.labels <- ids
    if (is.null(main)) main <- 
      paste("Prediction of ", aclass, ", GO-nodes: ", ncol(probs), sep="")
  } else {
    sens <- x@node.results$sensitivity[,aclass]
    spec <- x@node.results$specificity[,aclass]
    if (is.null(minspec)) {
      select <- (sens <= maxsens) & (sens >= minsens)
      minspec <- ifelse(sum(select)>50, sort(spec[select], decreasing=TRUE)[50], 0.0)
    }
    select <- (sens <= maxsens) & (sens >= minsens) & (spec >= minspec)
    if (sum(select) == 0) { warning("No nodes selected"); return(invisible(minspec)) }
    if (sum(select) == 1) {warning("Only one node selected");return(invisible(minspec))}
    trainset <- seq(dim(x@probs)[1])[-x@testset]
    if (dim(x@probs)[1] == length(x@cls)) report.samples <- which(x@cls==aclass)
    else {
      if (length(trainset) == 0) report.samples <- seq(dim(x@probs)[1])
      else {
        aclass.members <- trainset[which(x@cls==aclass)]
        report.samples <- unique(c(aclass.members,x@testset))
      }
    }
    probs <- x@probs[report.samples, aclass, select]
    spec <- round(100 * spec[select], 0)
    sens <- round(100 * sens[select], 0) 
    ids <- colnames(probs)
    if (full.names) left.labels <- paste(sens, "% / ", spec, "%", sep="")
    else left.labels <- paste(ids, " (", sens, "% / ", spec, "%)", sep="")
    if (is.null(main)) 
      main<-paste("Prediction of ", aclass, "\nSpecificity: [", minspec, ",1], ",
                  "Sensitivity [",minsens,",",maxsens,"], GO-nodes: ",sum(select),sep="")
  }
  if (!is.null(bin.thresh)) 
    probs<-matrix(as.numeric(bin.thresh<probs),ncol=ncol(probs),dimnames=dimnames(probs))

  # determine right labels if wanted
  right.labels <- NULL
  if (full.names) {
    for (i in 1:length(ids)) {
      rowid <- ids[i]
      id <- sub("a$", "", rowid)
      desc <- getGOterm(id)
      if(!is.null(desc)) {
        if (id != rowid) desc <- paste(desc, "(annot.)")
        if (nchar(desc) > max.label.length) 
          desc <- paste(substr(desc, 1, max.label.length-3), "...")
        right.labels <- c(right.labels, desc)
      } 
      else right.labels <- c(right.labels, id)
    }
  }

  # determine cluster orderings
  dy <- agnes(t(probs))
  dx <- agnes(probs)
  inv.dy <- rep(0, length(dy$order))
  for (i in 1:length(dy$order)) inv.dy[dy$order[i]] <- i
  inv.dx <- rep(0, length(dx$order))
  for (i in 1:length(dx$order)) inv.dx[dx$order[i]] <- i

  # determine widths of labels to optimize layout
  stam.opendev(outfile, "_pred_img", ps, res, pointsize)
  plot(c(0), type="n") # plot nothing to make strwidth happy
  left.label.widths  <- strwidth( left.labels, units="inches")
  right.label.widths <- strwidth(right.labels, units="inches")
  left.add <- ifelse(show.graph, 0.2, 0.5)
  margins <- c(1.0, max(left.label.widths) + left.add, 1.0, max(right.label.widths) + 0.5)
  height <- 2 + ncol(probs) * pointsize/72
  stam.closedev(outfile)
  leftsplit  <- 0.2
  rightsplit <- 0.9
  l <- ncol(probs)
  label.pos <- (seq(l)-1)/(l-1)

  # write client-side clickable map contencts
  xr <- round((ifelse(is.null(bin.thresh), rightsplit, 1) * width - 0.5) * res, 0)
  xl <- round(xr - max(right.label.widths)*res, 0)
  map <- ''
  for (i in 1:length(ids)) {
    y <- res + (length(ids)-i)*pointsize
    id <- ids[dy$order[i]]
    href <- sub(":", "", paste(id, ".html", sep=""))
    map <- paste(map, '<area shape="rect" href="', href, '" title="', id, 
                 '" target="NodeInfo" coords="', xl, ',', y, ',', xr,',',y+pointsize,'">',
                 sep="")
  }

  # draw graph
  stam.opendev(outfile, "_pred_img", ps, res, pointsize, height=height, width=width)
  if (invert) par(bg="black", fg="white", col.axis="white", col.main="white")
  if (show.graph) {
    par(mai=c(margins[1], 0.5, margins[3], 0.0), fig=c(0.0, leftsplit, 0.0, 1.0))
    stam.env[["root.dist"]] <- rep(0, length(ids))
    names(stam.env[["root.dist"]]) <- ids
    root.idx <- length(x@nodes)
    graph.lines <- collect.lines(root.idx, NULL, 0, which(select), x@nodes)
    tmp <- 0.5/(nrow(probs)-1)
    image(matrix(0,2,2), col="transparent", xlab="", ylab="", axes=FALSE,
         xlim=c(min(stam.env[["root.dist"]])-1, max(stam.env[["root.dist"]])+1), ylim=c(-tmp,1.0+tmp))
    points(stam.env[["root.dist"]], label.pos[inv.dy], pch=19)
    for (i in 1:nrow(graph.lines)) {
      pos <- graph.lines[i,]
      lines(stam.env[["root.dist"]][pos], label.pos[inv.dy[pos]])
    }
    mtext("GO-", side=1, line=0.5)
    mtext("Structure", side=1, line=1.5)
  }

  # draw the image
  if(show.graph) {
    if (!is.null(bin.thresh)) par(las=1, mai=margins, fig=c(leftsplit,1.0,0.0,1.0), new=TRUE)
    else par(las=1, mai=margins, fig=c(leftsplit, rightsplit,0.0, 1.0), new=TRUE)
  } else {
    if(!is.null(bin.thresh)) par(las=1, mai=margins)
    else par(las=1, mai=margins, fig=c(0, rightsplit,0.0, 1.0))
  }
  image(probs[dx$order, dy$order], axes=FALSE, zlim=c(0,1), main=main, col=col, ...)
  grid(nrow(probs), ncol(probs), lty=1, lwd=0.5, col="white")
  if (!is.null(left.labels)) axis(side=2, at=label.pos, labels=left.labels[dy$order])
  if ((full.names) && (length(x@cls) > 0)) 
    mtext("sens./spec.", side=1, line=1, at=-0.06, adj=1)
  if (!is.null(right.labels)) axis(side=4, at=label.pos, labels=right.labels[dy$order])
  box()
  l <- nrow(probs)
  label.pos <- (seq(l)-1)/(l-1)
  if (sample.labels) axis(side=1, at=label.pos, labels=rownames(probs)[dx$order], las=2)
  else {
    if (length(x@cls) > 0) {
      test.labels <- c('A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
                       'Q', 'R','S','T','U','V','W','X','Y','Z')
      xlabels <- rep("", nrow(probs))
      if (!identical(as.numeric(x@testset), as.numeric(1:length(x@cls)))) {
        test.positives <- report.samples[dx$order] %in% x@testset
        if (sum(test.positives) > length(test.labels)) xlabels[test.positives] <- '|'
        else xlabels[test.positives] <- test.labels[1:sum(test.positives)]
      }
      axis(side=1, at=label.pos, labels=xlabels, cex.axis=0.75)
    }
    if (is.null(xlab)) {
      if (length(x@cls) == 0) xlab <- paste(nrow(probs), "new samples")
      else {
        xlab <- paste(nrow(probs), " samples of class '", aclass, "'", sep="")
        if (!identical(as.numeric(x@testset), as.numeric(1:length(x@cls))))
          xlab <- paste(xlab, ", ", sum(test.positives),
                        " in test set marked with letters", sep="")
      }
    }
    mtext(xlab, side=1, line=2.5)
  }

  # color bar
  if (is.null(bin.thresh)) {
    par(mai=c(1.0, 0.0, 1.0, 0.8), fig=c(rightsplit, 1, 0.0, 1.0), new=TRUE)
    image(matrix(seq(0, 1, 1/(length(col)-1)), nrow=1), axes=FALSE, col=col)
    box()
    axis(side=4, at=seq(0, 1,0.2))
    mtext("Classifier Output", side=1, line=1, adj=0)
    mtext("colorcode", side=1, line=2, adj=0)  
  }
  if (!is.null(outfile)) dev.off()
  if (ps) return(invisible(minspec))
  else invisible(map)
}

plot.stamPrediction <- function(x, outfile=NULL, aclass=NULL, main=NULL,
                                minspec=0.9, minsens=0.1, maxsens=1.0,
                                res=72, ps=FALSE, pointsize=12, ...) {

  # nodewise scatter-plots
  if (length(x@cls) > 0) {
    stam.opendev(outfile, "_prednodes_plot", ps, res, pointsize)
    stam.nodes.plot(x, aclass=aclass, minspec=minspec, minsens=minsens, maxsens=maxsens)
    if (!is.null(outfile)) dev.off()
  }
  invisible(NULL)
}

#
#   End of file
#
################################################################################

