/* 
 * stamAux.c
 *
 * Auxiliary methods to accelerate StAM
 * 
 * written by Claudio Lottaz (MPI-MG) in October 2003
 */

#include <R.h>
#include <Rinternals.h>
#include <R_ext/Rdynload.h>
#include <math.h>

/* declare and register methods */

SEXP sumprobs(SEXP probs, SEXP children, SEXP weights);
SEXP redundancy(SEXP probs, SEXP nznodes);

void R_init_mylib(DllInfo *dllinfo)
{
  R_CallMethodDef callMethods[] = {
    {"sumprobs",   (void *)&sumprobs,   3},
    {"redundancy", (void *)&redundancy, 2},
    {NULL, NULL, 0}
  };
  R_registerRoutines(dllinfo, NULL, callMethods, NULL, NULL);
}

/*
 * sumprobs
 * 
 * replace the following R-code:
 *
 * > sum.probs <- function(probs, children, weights) {
 * >   newprobs <- matrix(0, nrow=dim(probs)[1], ncol=dim(probs)[2])
 * >   for (c in 1:length(children)) newprobs <- newprobs + weights[c] * probs[,,children[c]]
 * >   return(newprobs)
 * > }
 * >
 * > probs[,,n] <- sum.probs(probs, .stam.cv@nodes[[n]]@children, .stam.cv@nodes[[n]]@weights)
 *
 * replaced by:
 *
 * > probs[,,n] <- .Call("sumprobs", probs, children, weights)
 */

SEXP sumprobs(SEXP probs, SEXP children, SEXP weights)
{
  SEXP    probs_dims, newprobs;
  int     c, v, poffset, *xchildren;
  int     nnodes, nsamples, nclasses, nchildren, nvalues;
  double *xprobs, *xweights, *xnewprobs;

  PROTECT(probs    = coerceVector(probs,    REALSXP));
  PROTECT(children = coerceVector(children, INTSXP));
  PROTECT(weights  = coerceVector(weights,  REALSXP));

  if (!isReal(probs))       error("'probs' is not in real mode");
  if (!isInteger(children)) error("'children' is not in integer mode");
  if (!isReal(weights))     error("'weights' is not in real mode");
  if (!isArray(probs))      error("'children' is not a vector");
  if (!isVector(children))  error("'children' is not a vector");
  if (!isVector(weights))   error("'weights' is not a vector"); 

  nchildren = length(children);
  if (nchildren != length(weights)) error("'children' and 'weights' have unequal length");

  PROTECT(probs_dims = getAttrib(probs, R_DimSymbol));
  if (length(probs_dims) != 3) error("'probs' has not 3 dimensions");
  nsamples = INTEGER(probs_dims)[0];
  nclasses = INTEGER(probs_dims)[1];
  nnodes   = INTEGER(probs_dims)[2];
  nvalues  = nsamples * nclasses;

  PROTECT(newprobs = allocMatrix(REALSXP, nsamples, nclasses));

  xchildren   = INTEGER(children);
  xweights    = REAL(weights);
  xprobs      = REAL(probs);
  xnewprobs   = REAL(newprobs);

  for (v = 0; v < nvalues; v++) xnewprobs[v] = 0;
  for (c = 0; c <  nchildren; c++) {
    poffset = (xchildren[c] - 1) * nvalues;    
    for (v = 0; v < nvalues; v++) {
      xnewprobs[v] += xweights[c] * xprobs[poffset];
      poffset++;
    }
  }

  UNPROTECT(5);
  return(newprobs);
}

/*
 * redundancy
 *
 * replace the following R-code:
 *
 * > stam.redundancy <- function(p, nz.nodes=1:dim(p)[3]) {
 * >   # expects probability matrix p[samples, classes, nodes]
 * >   # only nodes in nz.nodes are computed
 * >   # returns a redundancy for each node and an average
 * >
 * >   pdiff <- function(p) 1 - sum(p[,1] * p[,2])
 * > 
 * >   nnodes <- length(nz.nodes)
 * >   nclasses <- dim(p)[2]
 * >   r <- rep(0, nnodes)
 * >   nz.nodes <- sort(nz.nodes)
 * >   for (i in 1:(nnodes-1)) {
 * >     for (j in (i+1):nnodes) {
 * >       tmp <- -sum(log(apply(p[,,c(nz.nodes[i],nz.nodes[j])], 1, pdiff)))
 * >       r[i] <- r[i] + tmp
 * >       r[j] <- r[j] + tmp
 * >     }
 * >   }
 * >   r <- r / (nnodes - 1) / dim(p)[1]
 * >   return(list(r = r, avg = sum(r)/nnodes))
 * > }
 * >
 * > r <- stam.redundancy(.stam.cv@probs, nz$nodes)
 *
 * replace by
 *
 * > r <- .Call("redundancy", .stam.cv@probs, nz$nodes)
 */

SEXP redundancy(SEXP probs, SEXP nznodes)
{
  SEXP    probs_dims, r, avg, ans, names;
  int     i, j, ci, cj, s, ioffset, joffset, *xnznodes;
  int     nnodes, nsamples, nclasses, nnznodes, nvalues;
  double *xprobs, *xr, *xavg, pdiff;

  PROTECT(probs    = coerceVector(probs,    REALSXP));
  PROTECT(nznodes  = coerceVector(nznodes,  INTSXP));

  if (!isReal(probs))       error("'probs' is not in real mode");
  if (!isInteger(nznodes))  error("'nznodes' is not in integer mode");
  if (!isArray(probs))      error("'children' is not a vector");
  if (!isVector(nznodes))   error("'children' is not a vector");

  nnznodes = length(nznodes);

  PROTECT(probs_dims = getAttrib(probs, R_DimSymbol));
  /*if (length(probs_dims) != 3) error("'probs' has not 3 dimensions");*/
  nsamples = INTEGER(probs_dims)[0];
  nclasses = INTEGER(probs_dims)[1];
  nnodes   = INTEGER(probs_dims)[2];
  nvalues  = nsamples * nclasses;

  PROTECT(r  = allocVector(REALSXP, nnznodes));
  PROTECT(avg = allocVector(REALSXP, 1));

  xnznodes    = INTEGER(nznodes);
  xprobs      = REAL(probs);
  xr          = REAL(r);
  xavg        = REAL(avg);

  for (i = 0; i < nnznodes; i++) xr[i] = 0;
  for (i = 0; i <  nnznodes - 1; i++) {
    ioffset = (xnznodes[i] - 1) * nvalues;    
    for (j = i + 1; j <  nnznodes; j++) {
      joffset = (xnznodes[j] - 1) * nvalues;    
      for (s = 0; s < nsamples; s++) {
	pdiff = 0;
	for (ci = 0; ci < nclasses; ci++) {
	  for (cj = 0; cj < nclasses; cj++) {
	    if (ci != cj) 
	      pdiff += xprobs[ioffset+ci*nsamples + s] * xprobs[joffset+cj*nsamples + s];
	  }
	}
	pdiff = log(pdiff);
	xr[i] -= pdiff;
	xr[j] -= pdiff;
      }
    }
  }
  xavg[0] = 0;
  nvalues = nsamples * (nnznodes - 1);
  for (i = 0; i < nnznodes; i++) {
    xr[i] /= nvalues;
    xavg[0] += xr[i];
  }
  xavg[0] /= nnznodes;

  /* return list with names r and avg */
  PROTECT(ans = allocVector(VECSXP, 2));
  SET_VECTOR_ELT(ans, 0, r);
  SET_VECTOR_ELT(ans, 1, avg);
  PROTECT(names = allocVector(STRSXP,2));
  SET_STRING_ELT(names, 0, mkChar("r"));
  SET_STRING_ELT(names, 1, mkChar("avg"));
  setAttrib(ans, R_NamesSymbol, names);
  UNPROTECT(7); 
  return(ans);
}

/*
 * End of file
 */

